#include "LargeAsmInt.h"

uint8_t	LargeAsmInt::htoi(char val)
{
	switch(val)
	{
	case '1':
		return 1;
	case '2':
		return 2;
	case '3':
		return 3;
	case '4':
		return 4;
	case '5':
		return 5;
	case '6':
		return 6;
	case '7':
		return 7;
	case '8':
		return 8;
	case '9':
		return 9;
	case 'a':
	case 'A':
		return 10;
	case 'b':
	case 'B':
		return 11;
	case 'c':
	case 'C':
		return 12;
	case 'd':
	case 'D':
		return 13;
	case 'e':
	case 'E':
		return 14;
	case 'f':
	case 'F':
		return 15;
	default:
		return 0;
	}
}

LargeAsmInt& LargeAsmInt::operator = (LargeAsmInt& inOther)
{
	uint32_t	*thisData = this->data,
				*otherData = inOther.data;

	__asm
	{
		mov ecx, 0
		mov edi, thisData
		mov esi, otherData

		cmp edi, esi
		je after_loop

copy_loop:
		mov eax, [esi+ecx*4]
		mov [edi+ecx*4], eax

		inc ecx
		cmp ecx, _LARGEASMINT_LENGTH_
		jl copy_loop
after_loop:
	}

	return *this;
};

LargeAsmInt& LargeAsmInt::operator = (const char* inVal)
{
	// check prefix
	if (strncmp(inVal, "0x", 2))
		return *this;

	null();

	int len = strlen(inVal) - 1;
	for (int i = 0; i < _LARGEASMINT_LENGTH_ && len > 0; ++i)
	{
		for (int x = 0; x < 8 && len > 0; ++x)
		{
			data[i] += htoi( inVal[len] ) << (x * 4);
			--len;
		}
	}

	return *this;
};

LargeAsmInt LargeAsmInt::operator+ (LargeAsmInt& inOther)
{
	LargeAsmInt result;
	uint32_t	*resultData = result.data,
				*thisData = this->data,
				*otherData = inOther.data;

	__asm
	{
		xor ecx, ecx						// loop counter
		mov edi, thisData					// this pointer
		mov esi, otherData					// inOther reference
		mov edx, resultData
		mov ebx, _LARGEASMINT_LENGTH_

		// according to measurements using this
		// 16bit split is faster than using adc
		movzx eax, WORD PTR [esi+ecx*4]
		movzx ebx, WORD PTR [edi+ecx*4]
		add eax, ebx
		mov [edx+ecx*4], ax
		shr eax, 16

		movzx ebx, WORD PTR [esi+ecx*4+2]
		add eax, ebx
		movzx ebx, WORD PTR [edi+ecx*4+2]
		add eax, ebx
		mov [edx+ecx*4+2], ax
		shr eax, 16

		inc ecx
		cmp ecx, _LARGEASMINT_LENGTH_
		jge after_plus_loop

before_plus_loop:
		movzx ebx, WORD PTR [esi+ecx*4]
		add eax, ebx
		movzx ebx, WORD PTR [edi+ecx*4]
		add eax, ebx
		mov [edx+ecx*4], ax
		shr eax, 16

		movzx ebx, WORD PTR [esi+ecx*4+2]
		add eax, ebx
		movzx ebx, WORD PTR [edi+ecx*4+2]
		add eax, ebx
		mov [edx+ecx*4+2], ax
		shr eax, 16

		inc ecx
		cmp ecx, _LARGEASMINT_LENGTH_
		jl before_plus_loop
after_plus_loop:
	}

	return result;
}

LargeAsmInt& LargeAsmInt::operator+= (LargeAsmInt& inOther)
{
	uint32_t	*thisData = this->data,
				*otherData = inOther.data;

	__asm
	{
		mov	ecx, 0							// loop counter
		mov edi, thisData					// this pointer
		mov esi, otherData					// inOther reference
		mov ebx, _LARGEASMINT_LENGTH_
		
		mov eax, [esi+ecx*4]
		add [edi+ecx*4], eax

		inc ecx
		dec ebx
		inc ebx
		dec ebx
//		jz stjsrjydf

/*before_loop:
		mov eax, [esi+ecx*4]
		adc [edi+ecx*4], eax

		inc ecx
		dec ebx
		jnz before_loop*/
stjsrjydf:
	}

	return *this;
};

LargeAsmInt LargeAsmInt::operator<< (uint32_t inOther)
{
	LargeAsmInt result;
	uint32_t	*thisData	= this->data;
	uint32_t	*resultData = result.data;
	uint32_t	limit;
	__asm
	{
		// pre-calculations
		mov limit, _LARGEASMINT_LENGTH_
		mov edx, 0
		mov eax, inOther
		mov ebx, 32
		div bx

		// return if all will become 0
		cmp eax, _LARGEASMINT_LENGTH_
		jge all_end		

		// calculate limit
		sub limit, eax

		// setup the memory pointers
		mov esi, resultData			// result data
		shl eax, 2					// go to the offset
		mov edi, thisData			// this data
		add esi, eax

		// check if we just need to copy
		cmp edx, 0
		je copy_loop
		
		xor edx, edx				// counter

		// calculate shifting values
		mov cl, dl					// remainder
		mov ch, 32
		sub ch, cl					// comlement remainder
			
		// first iteration
		mov eax, [edi+edx*4]
		mov ebx, eax
		shl eax, cl
		mov [esi+edx*4], eax

		inc edx
		cmp edx, limit
		jge all_end

loop_start:
		mov eax, [edi+edx*4]
		shl eax, cl
		xchg cl, ch
		shr ebx, cl
		or eax, ebx
		mov ebx, [edi+edx*4]
		xchg cl, ch
		mov [esi+edx*4], eax

		inc edx
		cmp edx, limit
		jl loop_start
		jmp all_end

copy_loop:
		mov eax, [edi+edx*4]
		mov [esi+edx*4], eax

		inc edx
		cmp edx, limit
		jl copy_loop
all_end:
	}

	return result;
};

LargeAsmInt LargeAsmInt::operator* (LargeAsmInt& inOther)
{
	LargeAsmInt result, tmp;
	uint32_t	*resultData = result.data,
				*thisData	= this->data,
				*otherData	= inOther.data;

	// get the address of the shifting operation
	LargeAsmInt (LargeAsmInt::*myshl)(uint32_t inOther);
	myshl = &LargeAsmInt::operator <<;

	// get the address of the addition +=
	LargeAsmInt& (LargeAsmInt::*myaddeq)(LargeAsmInt& inOther);
	myaddeq = &LargeAsmInt::operator +=;

	__asm
	{
		mov edi, otherData

		mov esi, 0
		mov ebx, 0
	
before_outer_loop:
		
		clc
		mov edx, 1						// set bitmask to 0x0001
before_inner_loop:
		mov eax, [edi+ebx*4]			// get the data
		and eax, edx					// is this bit set?
		jz after_calculation

		// shifting
		push edx						// save the bitmask
		push esi
		lea ecx, tmp
		push ecx						// return value
		mov ecx, this
		call DWORD PTR [myshl]

		// addition
		lea ecx, tmp
		push ecx						// parameter
		lea ecx, result
		call DWORD PTR [myaddeq]
		pop edx							// restore the bitmask

after_calculation:
		inc esi
		rcl edx, 1						// circle to the left
		jnc before_inner_loop			// until the 1 overflows, loop

		inc ebx							// outer loop
		cmp ebx, _LARGEASMINT_LENGTH_
		jl before_outer_loop
	}

	return result;
};